import click

from midi_utils.constants import CHORDS
from midi_utils.arp import STYLES
from saysynth.constants import (
    SAY_COLORS,
    SAY_EMPHASIS,
    SAY_ALL_PHONEMES,
    SAY_TUNED_VOICES,
    SAY_PHONEME_CLASSES,
    SAY_SEGMENT_MAX_DURATION,
    SAY_VOLUME_RANGE,
)


def group_options(*options):
    def wrapper(function):
        for option in reversed(options):
            function = option(function)
        return function

    return wrapper


def prepare_options_for_say(text: str, **kwargs):
    # handle some param edge cases
    rp = kwargs.get("randomize_phoneme")
    # for convenience, set the voice option to the one specified
    # in randomize phoneme.
    if rp and ":" in rp:
        kwargs["voice"] = rp.split(":")[0].strip().title()
    kwargs["text"] = text
    kwargs["use_tempfile"] = True
    return kwargs


# Duration Options

duration_opt = click.option(
    "-d",
    "--duration",
    default=None,
    type=int,
    help="The duration of the note in milliseconds.",
)

bpm_opt = click.option(
    "-db",
    "--duration-bpm",
    "duration_bpm",
    default=120,
    type=float,
    help="The bpm to use when calculating note duration.",
)
count_opt = click.option(
    "-dc",
    "--duration-count",
    "duration_count",
    default=4,
    type=str,
    show_default=True,
    help="The note length to use when calculating note duration (eg: 1/8 or 0.123 or 3)",
)
time_sig_opt = click.option(
    "-dts",
    "--duration-time-sig",
    "duration_time_sig",
    default="4/4",
    type=str,
    show_default=True,
    help="The time signature to use when calculating note duration",
)

duration_opts = group_options(duration_opt, bpm_opt, count_opt, time_sig_opt)

# Phoneme Options

phoneme_opt = click.option(
    "-p",
    "--phoneme",
    default="m",
    help="The phoneme to use.",
    show_default=True,
    type=click.Choice(SAY_ALL_PHONEMES),
)
randomize_phoneme_opt = click.option(
    "-rp",
    "--randomize-phoneme",
    show_default=True,
    help=(
        "Randomize the phoneme for every note. "
        "If `all` is passed, all phonemes will be used. "
        "Alternatively pass a list of phonemes (eg 'm,l,n') or a voice and style, eg: Fred:drone. "
        f"Valid voices include: {', '.join(SAY_TUNED_VOICES)}. "
        f"Valid styles include: {', '.join(SAY_PHONEME_CLASSES)}."
    ),
)

phoneme_opts = group_options(
    phoneme_opt,
    randomize_phoneme_opt,
)

# Start Options


randomize_start_opt = click.option(
    "-rt",
    "--randomize-start",
    type=int,
    nargs=2,
    help="Randomize the number of milliseconds to silence to add before the say text. The first number passed in is the minimum of the range, the second is the max.",
)
start_opt = click.option(
    "-t",
    "--start",
    default=None,
    show_default=True,
    type=float,
    help="The number of milliseconds of silence to add before the say text.",
)
start_bpm_opt = click.option(
    "-tb",
    "--start-bpm",
    default=120,
    type=float,
    help="The bpm to use when calculating start time",
)
start_count_opt = click.option(
    "-tc",
    "--start-count",
    default=0,
    type=str,
    show_default=True,
    help="The note length to use when calculating start time (eg: 1/8 or 0.123 or 3)",
)
start_time_sig_opt = click.option(
    "-tts",
    "--start-time-sig",
    default="4/4",
    type=str,
    show_default=True,
    help="The time signature to use when calculating start time",
)

start_opts = group_options(
    randomize_start_opt,
    start_opt,
    start_bpm_opt,
    start_count_opt,
    start_time_sig_opt,
)


# Segment Options

randomize_segments_opt = click.option(
    "-rs",
    "--randomize-segments",
    type=lambda x: [str(s) for s in x.split(",")] if len(x) else [],
    required=False,
    default=[],
    help="Randomize every segment's 'phoneme', 'note', and/or 'velocity'. Use commas to separate multiple randomization strategies",
)

segment_duration_opt = click.option(
    "-sd",
    "--segment-duration",
    default=SAY_SEGMENT_MAX_DURATION,
    show_default=True,
    type=lambda x: float(x) if x else None,
    help="The duration an individual phoneme",
)
segment_bpm_opt = click.option(
    "-sb",
    "--segment-bpm",
    default=120.0,
    type=float,
    help="The bpm to use when calculating phoneme duration",
)
segment_count_opt = click.option(
    "-sc",
    "--segment-count",
    default=1 / 16,
    type=str,
    show_default=True,
    help="The note length to use when calculating phoneme duration (eg: 1/8 or 0.123 or 3)",
)
segment_time_sig_opt = click.option(
    "-sts",
    "--segment-time-sig",
    default="4/4",
    type=str,
    show_default=True,
    help="The time signature to use when calculating phoneme duration",
)

segment_opts = group_options(
    randomize_segments_opt,
    segment_duration_opt,
    segment_bpm_opt,
    segment_count_opt,
    segment_time_sig_opt,
)

# Velocity Options

velocity_opt = click.option(
    "-vl",
    "--velocity",
    type=int,
    show_default=True,
    default=110,
    help="The midi velocity value to use for each note.",
)
velocity_emphasis_opt = click.option(
    "-ve",
    "--velocity-emphasis",
    "emphasis",
    type=int,
    nargs=2,
    show_default=True,
    default=SAY_EMPHASIS,
    help="Two midi velocity values (between 0 and 127) at which to add emphasis to a note/segment",
)
volume_range_opt = click.option(
    "-vr",
    "--volume-range",
    type=float,
    nargs=2,
    show_default=True,
    default=SAY_VOLUME_RANGE,
    help="The min and max volumes (range: 0.0-1.0) to use when mapping from midi velocities",
)
randomize_velocity_opt = click.option(
    "-rv",
    "--randomize-velocity",
    type=int,
    nargs=2,
    help="Randomize a note's velocity by supplying a min and max midi velocity (eg: -rv 40 120)",
)

velocity_opts = group_options(
    velocity_opt,
    velocity_emphasis_opt,
    volume_range_opt,
    randomize_velocity_opt,
)


# ADSR Options

attack_opt = click.option(
    "-at",
    "--attack",
    default=0.0,
    show_default=True,
    type=float,
    help="The percentage of the duration it takes to reach the max volume of the note",
)
decay_opt = click.option(
    "-de",
    "--decay",
    default=0.0,
    type=float,
    help="The percentage of the duration it takes to reach the sustain volume of the note",
)
sustain_opt = click.option(
    "-su",
    "--sustain",
    default=1.0,
    type=float,
    show_default=True,
    help="The the sustain volume of the note",
)
release_opt = click.option(
    "-re",
    "--release",
    default=0.0,
    type=float,
    show_default=True,
    help="The percentage of the duration it takes to reach the min volume of the note",
)

adsr_opts = group_options(
    attack_opt,
    decay_opt,
    sustain_opt,
    release_opt,
)

# Say Options
exec_opt = click.option(
    "-x",
    "--exec",
    is_flag=True,
    default=False,
    help="Run the generated text through the say command.",
)

rate_opt = click.option(
    "-r", "--rate", type=int, default=70, show_default=True, help="Rate to speak at"
)
voice_opt = click.option(
    "-v",
    "--voice",
    type=click.Choice(SAY_TUNED_VOICES),
    default="Fred",
    show_default=True,
    help="Voice to use",
)
# input_file_opt = click.option(
#     '-i',
#     "--input-file",
#     type=click.File(mode = "w"),
#     help="File to read text input from"
# )
output_file_opt = click.option(
    "-ao",
    "--audio-output-file",
    type=str,
    help="File to write audio output to",
)
audio_device_opt = click.option(
    "-ad",
    "--audio-device",
    type=str,
    help="Name of the audio device to send the signal to",
)
networks_send_opt = click.option(
    "-ns", "--network-send", type=str, help="Network address to send the signal to"
)
stereo_opt = click.option(
    "-st",
    "--stereo",
    is_flag=True,
    default=False,
    help="Whether or not to generate a stereo signal",
)
sample_size_opt = click.option(
    "-ss",
    "--sample-size",
    type=int,
    default=32,
    help="Sample size of the signal (1:32)",
)
sample_rate_opt = click.option(
    "-sr",
    "--sample-rate",
    type=int,
    default=22050,
    help="Sample rate of the signal (0:22050)",
)
quality_opt = click.option(
    "-qu", "--quality", type=int, default=127, help="Quality of the signal (1:127)"
)
progress_bar_opt = click.option(
    "-pg",
    "--progress",
    is_flag=True,
    default=False,
    help="Whether or not to display an interactive progress bar",
)
interactive_opt = click.option(
    "-in",
    "--interactive",
    is_flag=True,
    default=False,
    help="Whether or not to display highlighted text",
)
text_color_opt = click.option(
    "-cf",
    "--text-color",
    type=click.Choice(SAY_COLORS),
    default="white",
    help="The text color to use when displaying highlighted text",
)
bg_color_opt = click.option(
    "-cb",
    "--bg-color",
    type=click.Choice(SAY_COLORS),
    default="black",
    help="The background color to use when displaying highlighted text",
)


say_opts = group_options(
    exec_opt,
    rate_opt,
    voice_opt,
    output_file_opt,
    audio_device_opt,
    networks_send_opt,
    stereo_opt,
    sample_size_opt,
    sample_rate_opt,
    quality_opt,
    progress_bar_opt,
    interactive_opt,
    text_color_opt,
    bg_color_opt,
)

# Chord Options

chord_opt = click.option(
    "-c",
    "--chord",
    required=False,
    default="min6_9",
    type=click.Choice([c.lower() for c in CHORDS.keys()]),
    help="An optional name of a chord to build using the note as root.",
)
chord_notes_opt = click.option(
    "-cn",
    "--chord-notes",
    required=False,
    default=[],
    type=lambda x: [int(i.strip()) for i in x.split(",")] if x else [],
    help="An optional list of midi numbers to build a chord from the root.",
)
chord_velocities_opt = click.option(
    "-cv",
    "--chord-velocities",
    required=False,
    type=lambda x: [int(i.strip()) for i in x.split(",")] if x else [],
    help="A list of integers (eg: '50,100,127') specifying the midi velocity each note i the chord. The length of this list much match the number of notes in the chord. --volume-range anyd --velocity-steps also modify this parameter",
)
chord_inversions_opt = click.option(
    "-ci",
    "--chord-inversions",
    "inversions",
    default=[],
    required=False,
    type=lambda x: [int(i.strip()) for i in x.split(",")] if x else [],
    help="A list of integers (eg: '0,1,-1') specifying the direction and amplitude to invert each note. The length of this list much match the number of notes in the chord (post-stack).",
)
chord_stack_opt = click.option(
    "-cs",
    "--chord-stack",
    "stack",
    default=0,
    required=False,
    type=int,
    help="Stack a chord up (eg: '1' or '2') or down (eg: '-1' or '-2').",
)

chord_opts = group_options(
    chord_opt,
    chord_notes_opt,
    chord_inversions_opt,
    chord_stack_opt,
)

# Arp Options

#
notes_opt = click.option(
    "-ns",
    "--notes",
    required=False,
    default=[],
    type=lambda x: [i.strip() for i in x.split(",")] if x else [],
    help="a list of note names / midi note numbers to argpeggiate",
)

octaves_opt = click.option(
    "-oc",
    "--octaves",
    required=False,
    default=[],
    type=lambda x: [int(i.strip()) for i in x.split(",")] if x else [],
    help="a list of octaves to add to the notes",
)

styles_opt = click.option(
    "-sl",
    "--styles",
    required=False,
    default="down",
    type=lambda x: [str(i.strip().lower()) for i in x.split(",")] if x else [],
    help=f"A list of styles/sorting algorithms to apply to the notes. This occurs after octaves are added. \nchoose from:\n {', '.join([str(k) for k in STYLES.keys()])}",
)

velocities_opt = click.option(
    "-vl",
    "--velocities",
    required=False,
    default="100",
    show_default=True,
    type=lambda x: [int(i.strip()) for i in x.split(",")] if x else [],
    help="a list of velocities to apply to the notes, if this list is shorter than the list of notes, a modulo operator is performed.",
)

loops_opt = click.option(
    "-l",
    "--loops",
    default=4,
    show_default=True,
    type=int,
    help="The number of times to loop the notes",
)

## beat duration

beat_duration_opt = click.option(
    "-bd",
    "--beat-duration",
    default=None,
    required=False,
    type=int,
    help="The duration of the note in milliseconds.",
)
beat_bpm_opt = click.option(
    "-bb",
    "--beat-bpm",
    default=120,
    type=float,
    show_default=True,
    help="The bpm to use when calculating note duration. --duration",
)
beat_count_opt = click.option(
    "-bc",
    "--beat-count",
    default=4,
    type=str,
    show_default=True,
    help="The note length to use when calculating note duration (eg: 1/8 or 0.123 or 3)",
)
beat_time_sig_opt = click.option(
    "-bts",
    "--beat-time-sig",
    default="4/4",
    type=str,
    show_default=True,
    help="The time signature to use when calculating note duration",
)

## note duration

note_duration_opt = click.option(
    "-od",
    "--note-duration",
    default=None,
    required=False,
    type=int,
    help="The duration of a single note in the arp. Defaults to the beat duration",
)
note_bpm_opt = click.option(
    "-nb",
    "--note-bpm",
    "note_bpm",
    default=120,
    type=float,
    show_default=True,
    help="The bpm to use when calculating note duration. --duration",
)
note_count_opt = click.option(
    "-nc",
    "--note-count",
    "note_count",
    default=4,
    type=str,
    show_default=True,
    help="The note length to use when calculating note duration (eg: 1/8 or 0.123 or 3)",
)
note_time_sig_opt = click.option(
    "-nts",
    "--note-time-sig",
    default="4/4",
    type=str,
    show_default=True,
    help="The time signature to use when calculating note duration",
)


arp_opts = group_options(
    notes_opt,
    octaves_opt,
    styles_opt,
    velocities_opt,
    loops_opt,
    beat_duration_opt,
    beat_bpm_opt,
    beat_count_opt,
    beat_time_sig_opt,
    note_duration_opt,
    note_bpm_opt,
    note_count_opt,
    note_time_sig_opt,
)



import sys

import click

from saysynth import MidiTrack
from saysynth.cli.options import (
    phoneme_opts,
    segment_opts,
    start_opts,
    adsr_opts,
    say_opts,
    velocity_emphasis_opt,
    volume_range_opt,
    randomize_velocity_opt,
    prepare_options_for_say,
    group_options
)
from saysynth.lib import say

# # # ##
# Midi #
# # # ##

def run(**kwargs):
    """
    Given a midi file, generate a text-file (or stdout stream) of phonemes with pitch information for input to say
    """

    # generate the text
    text = MidiTrack(**kwargs).to_say_text()

    # handle writing text to file
    output_file = kwargs.get("output_file")
    if output_file:
        with open(output_file, "w") as f:
            f.write(text)
        sys.exit(0)

    # if we're not executing say, write text to stdout
    if not kwargs.get('exec'):
        sys.stdout.write(text)
        sys.exit(0)

    # execute say
    say.run(**prepare_options_for_say(text, **kwargs))



@click.command()
@click.argument("midi_file", required=True)
@click.option(
    "-l",
    "--loops",
    default=1,
    show_default=True,
    type=int,
    help="The number of times to loop the midi file",
)
@start_opts
@phoneme_opts
@group_options(velocity_emphasis_opt, volume_range_opt, randomize_velocity_opt)
@adsr_opts
@segment_opts
@click.option(
    "-o",
    "--output-file",
    type=str,
    help="A filepath to write the generated text to",
)
@say_opts
def cli(**kwargs):
    return run(**kwargs)


import click
import os

COMMAND_METHOD = "cli"
COMMANDS_FOLDER = os.path.join(os.path.dirname(__file__), "commands")
COMMAND_NAMES = [
    filename[:-3]
    for filename in os.listdir(COMMANDS_FOLDER)
    if filename.endswith(".py") and not filename.startswith("__")
]


class SaySynthCLI(click.MultiCommand):
    commands = COMMAND_NAMES

    def list_commands(self, ctx):
        return self.commands

    def get_command(self, ctx, name):
        if name not in self.commands:
            raise ValueError(f"Invalid command name: {name}")
        ns = {}
        fn = os.path.join(COMMANDS_FOLDER, name + ".py")
        with open(fn) as f:
            code = compile(f.read(), fn, "exec")
            eval(code, ns, ns)
        return ns[COMMAND_METHOD]


main = SaySynthCLI(help="Turn the `say` command into a synthesizer")

if __name__ == "__main__":
    main()



import os
import math
import tempfile
from typing import Union

# midi / note / freq lookup


def frange(start, stop, by, sig_digits=5):
    """
    Generate a range of float values.
    """
    div = math.pow(10, sig_digits)
    for value in range(int(start * div), int(stop * div), int(by * div)):
        yield round(value / div, sig_digits)


def here(f, *args):
    """
    Pass __file__ to get the current directory and *args to generate a filepath
    """
    return os.path.join(os.path.dirname(os.path.abspath(f)), *args)


def make_tempfile(format: str = "txt"):
    """
    Make a tempfile
    """
    return tempfile.mkstemp(suffix=f".{format}")[-1]


def bpm_to_time(
    bpm: float = 120.00, count: Union[str, int, float] = 1, timesig: str = "4/4"
):
    """
    Take a bpm, note count, and timesig and return a length in seconds
    """
    if isinstance(count, str):
        if "/" in count:
            numerator, denominator = count.split("/")
            count = float(numerator) / float(denominator)
    time_segs = timesig.split("/")
    return (60.00 / float(bpm)) * float(time_segs[0]) * float(count) * 1000.0


def rescale(x, range_x, range_y, sig_digits=3):

    # Figure out how 'wide' each range is
    x_min = min(range_x)
    y_min = min(range_y)
    x_span = max(range_x) - x_min
    y_span = max(range_y) - y_min

    # Compute the scale factor between left and right values
    scale_factor = float(y_span) / float(x_span)

    return round(y_min + (x - x_min) * scale_factor, sig_digits)



from saysynth.utils import bpm_to_time


def test_bpm_to_time_str_count():
    t = bpm_to_time(60, "1/16")
    assert t == 250


def test_bpm_to_time_str_count_time_sig():
    t = bpm_to_time(60, "1/8", "3/3")
    assert t == 375


def test_bpm_to_time_float_count():
    t = bpm_to_time(60, 1.0 / 8.0, "3/3")
    assert t == 375



from typing import List, Tuple

import nltk

try:
    nltk.data.find("tokenizers/punkt")
except LookupError:
    nltk.download("cmudict")
try:
    nltk.data.find("tokenizers/punkt")
except LookupError:
    nltk.download("punkt")
from nltk.tokenize import word_tokenize
from nltk.corpus import cmudict
from g2p_en import G2p

from ..constants import G2P_PHONEMES_TO_SAY_PHONEMES


G2P = G2p()
CMU = cmudict.dict()


def word_to_g2p_phonemes(text: str) -> List[str]:
    return G2P(text)


def word_to_say_phonemes(text: str) -> List[str]:
    g2p_phonemes = word_to_g2p_phonemes(text)
    return [G2P_PHONEMES_TO_SAY_PHONEMES.get(p, "") for p in g2p_phonemes]


def word_to_syllable_count(word: str) -> int:
    try:
        return [len(list(y for y in x if y[-1].isdigit())) for x in CMU[word.lower()]][
            0
        ]
    except KeyError:
        # if word not found in cmudict
        # referred from stackoverflow.com/questions/14541303/count-the-number-of-syllables-in-a-word
        count = 0
        vowels = "aeiouy"
        word = word.lower()
        if word[0] in vowels:
            count += 1
        for index in range(1, len(word)):
            if word[index] in vowels and word[index - 1] not in vowels:
                count += 1
        if word.endswith("e"):
            count -= 1
        if word.endswith("le"):
            count += 1
        if count == 0:
            count += 1
        return count


def process_text_for_say(text) -> List[Tuple[int, List[str], str]]:
    """
    Get a list of phonemes + syllable counts for each word in a text
    """
    return [
        (word_to_syllable_count(word), word_to_say_phonemes(word), word)
        if word not in [",", ".", "?", "!", "-", ":", ";"]
        else (1, ["%"])  # silence
        for word in word_tokenize(text)
    ]



import random
import copy
from typing import Optional, List, Union, Tuple, Dict, Any

from midi_utils import (
    midi_arp,
    midi_to_freq,
    note_to_midi,
    midi_chord,
    midi_to_note,
    ADSR,
)

from .lib import midi, say, nlp
from .cli.options import prepare_options_for_say
from .utils import bpm_to_time, frange, rescale
from .constants import (
    SAY_PHONEME_SILENCE,
    SAY_SEGMENT_MAX_DURATION,
    SAY_SEGMENT_SILENCE_DURATION,
    SAY_EMPHASIS,
    SAY_PHONEME_CLASSES,
    SAY_ALL_PHONEMES,
    SAY_TUNE_TAG,
    SAY_TUNED_VOICES,
    SAY_PHONEME_VOICE_CLASSES,
    SAY_VOLUME_RANGE,
)


class Word(object):
    def __init__(self, *args):
        self.text = args[2]
        self.syllables = args[0]
        self.phonemes = args[1]


class Lyrics(object):
    # A lyric object is initialized from a text and containes info about the words in a song
    def __init__(self, text):
        self.text = text
        self.words = [Word(*word) for word in nlp.process_text_for_say(text)]


class Segment(object):
    # a segment represents an individaul part-of-speech (phoneme) in say
    # a segment can have a duration, a rate, and a sequences of pitches
    # its final state is a string for input to say
    # TODO: move randomization out of here and into the note class
    #
    def __init__(
        self,
        note: Union[int, str],
        velocity: int = 127,
        phoneme: str = "m",
        duration: Union[float, int] = SAY_SEGMENT_MAX_DURATION,
        type: str = "note",
        emphasis: Tuple[int, int] = SAY_EMPHASIS,
        volume_range: Tuple[float, float] = SAY_VOLUME_RANGE,
        duration_sig_digits: int = 4,
        **kwargs,
    ):
        self._phoneme = phoneme
        self._duration = duration
        self._emphasis = emphasis
        self.velocity = velocity
        self.note = note
        self.is_silence = type == "silence"
        self.volume_range = volume_range
        self.duration_sig_digits = duration_sig_digits

    @property
    def phoneme(self):
        return self._phoneme

    @property
    def frequency_envelope(self):
        # lookup frequency
        freq = midi_to_freq(note_to_midi(self.note))
        return f"P {freq}:0 {freq}:100"

    @property
    def duration(self):
        """
        calculate phoneme duration or pass it through
        """
        return round(
            min(self._duration, SAY_SEGMENT_MAX_DURATION), self.duration_sig_digits
        )

    @property
    def volume(self) -> str:
        """
        Translate a midi velocity value (0-127) into a pair of say volume tags, eg: "[[ volm +0.1 ]]"
        """
        return f"[[ volm {rescale(self.velocity, [0, 127], self.volume_range)} ]]"

    @property
    def emphasis(self) -> str:
        """
        Translate a midi velocity value (0-127) into a phoneme emphasis value ("", "1", or "2")
        when provided with a tuple of steps (step_1, step_2) eg: (75, 100)
        """
        if not self.velocity:
            return ""
        if self.velocity > self._emphasis[1]:
            return "2"
        if self.velocity > self._emphasis[0]:
            return "1"
        return ""

    def to_text(self):
        """ "
        Format phoneme text.
        TODO: Handle intra-note modulation?
        """
        if self.is_silence:
            return f"{self.volume} {SAY_PHONEME_SILENCE} {{D {self.duration}}}"
        return f"{self.volume} {self.emphasis}{self.phoneme} {{D {self.duration}; {self.frequency_envelope}}}"

    def __eq__(self, other):
        return str(self) == str(other)


class Note(object):
    def __init__(
        self,
        note: Union[int, str] = "A3",
        phoneme: str = "m",
        # start position
        start: Optional[int] = 0,
        start_bpm: Optional[Union[float, int]] = 120,
        start_count: Union[str, float, int] = 0,
        start_time_sig: str = "4/4",
        # envelope
        velocity: int = 127,
        attack: Union[float, int] = 0,
        decay: Union[float, int] = 0,
        sustain: Union[float, int] = 1,
        release: Union[float, int] = 0,
        # length
        duration: Optional[Union[float, int]] = None,
        duration_bpm: Optional[Union[float, int]] = 120,
        duration_count: Union[str, float, int] = 1,
        duration_time_sig: str = "4/4",
        # segmentation
        segment_duration: Optional[int] = None,
        segment_bpm: Optional[float] = 120,
        segment_count: Optional[Union[str, float, int]] = 1.0 / 8.0,
        segment_time_sig: Optional[str] = "4/4",
        # randomization
        randomize_phoneme: Optional[str] = None,
        randomize_velocity: Optional[Tuple[int, int]] = None,
        randomize_octave: Optional[List[int]] = [],
        randomize_segments: Optional[List[str]] = [],
        randomize_start: Optional[Tuple[int, int]] = None,
        **segment_options,
    ):
        """
        Generate say text for a collection of phonemes with adsr + pitch modulation
        """

        self.segment_options = segment_options
        root = segment_options.pop("root", None)

        if root or note:
            self.note = note_to_midi(root or note)  # root == note
            self.name = midi_to_note(self.note)
        else:
            self.note = 0
            self.name = "silence"

        # phoneme
        self.phoneme = phoneme

        # start position
        self.start = start
        if not self.start:
            self.start = bpm_to_time(start_bpm, start_count, start_time_sig)
        if randomize_start:
            self.start = random.choice(
                range(self.randomize_start[0], self.randomize_start[1] + 1)
            )

        # duration
        self.duration = duration
        if not self.duration:
            self.duration = bpm_to_time(duration_bpm, duration_count, duration_time_sig)

        # velocity
        self.velocity = velocity

        # segmentation
        self.segment_duration = segment_duration
        if not self.segment_duration:
            self.segment_duration = bpm_to_time(
                segment_bpm, segment_count, segment_time_sig
            )
        self.segment_duration = min(SAY_SEGMENT_MAX_DURATION, self.segment_duration)
        self.segment_count = int(self.duration / self.segment_duration) + 1
        self._segments = []

        # adsr
        self.adsr = ADSR(attack, decay, sustain, release, samples=self.segment_count)

        # randomization
        self.randomize_phoneme = randomize_phoneme
        self.randomize_velocity = randomize_velocity
        self.randomize_octave = randomize_octave
        self.randomize_segments = randomize_segments
        self.randomize_start = randomize_start

    def _get_phoneme(self):
        # handle phoneme randomization
        if self.randomize_phoneme:
            if self.randomize_phoneme == "all":
                return random.choice(SAY_ALL_PHONEMES)
            elif ":" in self.randomize_phoneme:
                voice, style = self.randomize_phoneme.split(":")
                voice = voice.title()  # allow for lowercase
                try:
                    return random.choice(SAY_PHONEME_VOICE_CLASSES[voice][style])
                except KeyError:
                    raise ValueError(
                        f"Invalid `voice` '{voice}' or `style` '{style}'. "
                        f"`voice` must be one of: {', '.join(SAY_TUNED_VOICES)}. "
                        f"`style` must be one of: {', '.join(SAY_PHONEME_CLASSES)}"
                    )
            else:
                return random.choice(
                    [c.strip() for c in self.randomize_phoneme.split(",")]
                )
        return self.phoneme

    def _get_note(self):
        if self.randomize_octave:
            return (random.choice(self.randomize_octave) * 12) + note_to_midi(self.note)
        return self.note

    def _get_velocity(self):
        if self.randomize_velocity:
            return random.choice(
                range(self.randomize_velocity[0], self.randomize_velocity[1] + 1)
            )
        return self.velocity

    def _get_segment_kwargs(self, **kwargs):
        opts = copy.copy(self.segment_options)
        opts.update(kwargs)
        return opts

    def _get_segment(
        self,
        index=0,
        note=None,
        velocity=0,
        phoneme=None,
        duration=None,
        **kwargs,
    ):
        # optionally randomize every segment.
        if "note" in self.randomize_segments and self.randomize_velocity:
            note = self._get_note()
        if "velocity" in self.randomize_segments and self.randomize_velocity:
            velocity = self._get_velocity()
        if "phoneme" in self.randomize_segments and self.randomize_phoneme:
            phoneme = self._get_phoneme()

        return Segment(
            note=note,
            velocity=velocity * self.adsr.get_value(index),
            phoneme=phoneme,
            duration=duration or self.segment_duration,
            **self._get_segment_kwargs(**kwargs),
        )

    @property
    def segments(self):

        # cache segments
        if len(self._segments) > 0:
            return self._segments

        # get initial value of note + velocity + phoneme
        note = self._get_note()
        velocity = self._get_velocity()
        phoneme = self._get_phoneme()

        if self.start and self.start > 0:
            # create multiple silent phonemes which add up to the desired start position
            start_breaks = list(
                frange(0.0, self.start, SAY_SEGMENT_SILENCE_DURATION, 10)
            )
            for index, total_start_time in enumerate(start_breaks[1:]):
                segment = self._get_segment(index, type="silence", velocity=0)
                self._segments.append(segment)

            if total_start_time < self.start:
                # add final step of silence
                self._segments.append(
                    self._get_segment(index + 1, type="silence", velocity=0)
                )

        # create multiple phonemes which add up to the phoneme_duration
        segment_breaks = list(frange(0.0, self.duration, self.segment_duration, 10))
        total_time = 0
        index = 0
        for index, total_time in enumerate(segment_breaks[1:]):
            segment = self._get_segment(
                index,
                note,
                velocity,
                phoneme,
                type=self.segment_options.get("type", "note"),
            )
            self._segments.append(segment)

        if total_time < self.duration and len(self._segments) < self.segment_count:

            # add final step
            self._segments.append(
                self._get_segment(
                    index + 1,
                    note,
                    velocity,
                    phoneme,
                    duration=self.duration - total_time,
                    type=self.segment_options.get("type", "note"),
                )
            )
        return self._segments

    def to_text(self):
        return "\n".join([s.to_text() for s in self.segments])

    def to_say_text(self):
        return SAY_TUNE_TAG + "\n" + self.to_text()

    def __eq__(self, other):
        return self.to_text() == other.to_text()


class Chord(object):
    _notes = []

    def __init__(self, **kwargs):
        og_note = kwargs.pop("note", "A1")
        self.root = kwargs.pop("root", og_note)  # root == note
        self.root_midi = note_to_midi(self.root)
        self.note_options = kwargs
        # check for arbitrary notes
        notes = kwargs.pop("chord_notes", [])
        if len(notes):
            self.midi_notes = [self.root_midi + n for n in notes]

        # otherwise use chord
        else:
            self.midi_notes = midi_chord(root=self.root, **self.note_options)

    def get_kwargs(self, **kwargs):
        """
        get kwargs + update with new ones
        used for mapping similar kwards over different notes
        """
        d = dict(self.note_options.items())
        d.update(kwargs)
        return d

    @property
    def notes(self):
        if not len(self._notes):
            for n in self.midi_notes:
                self._notes.append(Note(**self.get_kwargs(note=n)))
        return self._notes

    def write(self, output_file):
        for note in self.notes:
            fn = ".".join(output_file.split(".")[:-1])
            ext = "txt" if "." not in output_file else output_file.split(".")[-1]
            note_output_file = f"{fn}-{note.note}.{ext}"
            with open(note_output_file, "w") as f:
                f.write(str(note))

    def play(self):
        # generate a command for each note in the chord
        commands = []
        for note in self.notes:
            cmd_kwargs = prepare_options_for_say(
                text=note.to_say_text(), **self.get_kwargs(note=note.note, type="note")
            )
            commands.append(cmd_kwargs)
        say.spawn(commands)


class Arp(object):
    _notes = []

    def __init__(
        self,
        notes: List[int] = [],  # arbitrary list of notes to arpeggiate
        root: str = "A3",  # root note of chord
        chord: str = "min6_9",  # chord name,
        inversions: List[int] = [],  # inversions list
        stack: int = 0,  # stack a chord up or down
        styles: List[str] = ["down"],
        octaves: List[int] = [0],  # a list of octaves to add to the notes (eg: [-1, 2])
        velocities: List[int] = [100],
        # a list of velocities to apply across notes,
        # velocities are retrieved using a modulo so
        # this can be any duration and will be applied
        # in order
        beat_bpm: float = 131.0,  # bpm to use when determining beat length
        beat_count: Union[float, int, str] = "1/16",  # count of one beat of the arp
        beat_time_sig: str = "4/4",  # time signature of arp
        beat_duration: Optional[float] = None,  # the time of the note in ms
        note_bpm: float = 131.0,  # bpm to use when determining note duration
        note_count: Union[
            float, int, str
        ] = "3/64",  # count of note legth (should be less than or equat to beat count)
        note_time_sig: str = "4/4",  # time signature of arp
        note_duration: Optional[float] = None,  # arbitrary duration of note in ms
        randomize_start: Optional[List[int]] = None,
        start_bpm: float = 131.0,  # bpm to use when determining the start of the arp and adds silence at the beginning
        start_count: Union[float, int, str] = 0,  # the start beat count
        start_time_sig: str = "4/4",  # time signature to use when determining start
        start_duration: Optional[
            float
        ] = None,  # the amount of silence to add at the beginning in ms
        duration_bpm: float = 131.0,  # bpm to use when determining note duration
        duration_count: Union[float, int, str] = "16",  # the duration beat count
        duration_time_sig: str = "4/4",  # time signature to use when determining duration
        duration: Optional[float] = None,  # the total duration of the pattern in ms
        loops: Optional[int] = None,
        **note_options,
    ):
        self.styles = styles

        if randomize_start:
            start_duration = random.choice(
                range(randomize_start[0], randomize_start[1] + 1)
            )

        self.sequence = midi_arp(
            notes=notes,
            root=root,
            chord=chord,
            inversions=inversions,
            stack=stack,
            octaves=octaves,
            styles=styles,
            velocities=velocities,
            beat_bpm=beat_bpm,
            beat_count=beat_count,
            beat_time_sig=beat_time_sig,
            beat_duration=beat_duration,
            note_bpm=note_bpm,
            note_count=note_count,
            note_time_sig=note_time_sig,
            note_duration=note_duration,
            start_bpm=start_bpm,
            start_count=start_count,
            start_time_sig=start_time_sig,
            start_duration=start_duration,
            duration_bpm=duration_bpm,
            duration_count=duration_count,
            duration_time_sig=duration_time_sig,
            duration=duration,
            loops=loops,
        )
        self.note_options = note_options

    def get_kwargs(self, **kwargs):
        """
        get kwargs + update with new ones
        used for mapping similar kwards over different notes
        """
        d = dict(self.note_options.items())
        d.update(kwargs)
        return d

    @property
    def notes(self):
        if not len(self._notes):
            for note in self.sequence:
                self._notes.append(Note(**self.get_kwargs(**note)))
        return self._notes

    def to_text(self):
        return "\n".join([n.to_text() for n in self.notes])

    def to_say_text(self):
        return SAY_TUNE_TAG + "\n" + self.to_text()

    def __repr__(self):
        return f"<Arp {','.join(self.styles)}"


class MidiTrack(object):
    def __init__(
        self,
        midi_file: str,
        loops: int = 1,
        # start position
        start: Optional[int] = None,
        start_bpm: Optional[Union[float, int]] = 120,
        start_count: Union[str, float, int] = 0,
        start_time_sig: str = "4/4",
        **kwargs,
    ):
        """
        TODO: multichannel midi
        """
        self.midi_file = midi_file
        self.loops = loops

        self.start = start
        if not self.start:
            self.start = bpm_to_time(start_bpm, start_count, start_time_sig)
        self.start_segment_count = int(self.start / SAY_SEGMENT_SILENCE_DURATION) + 1
        self.kwargs = kwargs
        self._notes = []
        self._start_segments = []

    @property
    def notes(self):
        if len(self._notes):
            return self._notes
        for _ in range(0, self.loops):
            for note in midi.process(self.midi_file):
                self.kwargs.update(note)
                self._notes.append(Note(**self.kwargs))
        return self._notes

    def _get_start_text(self):
        if not self.start:
            return ""

        # create multiple silent segments which add up to the start position
        if not len(self._start_segments):
            for _, total_time in enumerate(
                list(frange(0.0, self.start, SAY_SEGMENT_SILENCE_DURATION, 10))[1:]
            ):
                self._start_segments.append(
                    Segment(
                        type="silence",
                        velocity=0,
                        duration=SAY_SEGMENT_SILENCE_DURATION,
                    )
                )

            if (
                total_time < self.start
                and len(self._start_segments) < self.start_segment_count
            ):
                # add final silent step
                self._start_segments.append(
                    Segment(
                        type="silence", velocity=0, duration=self.start - total_time
                    )
                )
        return "\n".join(self._start_segments)

    def to_text(self):
        return self._get_start_text() + "\n" + "\n".join(map(str, self.notes))

    def to_say_text(self):
        return SAY_TUNE_TAG + "\n" + self.to_text()

    def __repr__(self):
        return f"<MidiTrack {self.midi_file}>"



from saysynth.utils import here
from saysynth.lib import nlp
from saysynth.constants import G2P_PHONEMES_TO_SAY_PHONEMES


def test_g2p_phonemes_to_say_phonemes():
    g2p_texts = nlp.word_to_g2p_phonemes("hello world")
    for t in g2p_texts:
        if t.strip():
            assert t in G2P_PHONEMES_TO_SAY_PHONEMES
    say_texts = nlp.word_to_say_phonemes("hello world")
    for t in say_texts:
        if t.strip():
            assert t in set(G2P_PHONEMES_TO_SAY_PHONEMES.values())


def test_word_to_syllable_count():
    word = "hello"
    syllables = nlp.word_to_syllable_count(word)
    assert syllables == 2


def test_word_to_syllable_count():
    text = "hello world. I am a computer."
    words = nlp.process_text_for_say(text)
    assert words[0][1] == ["h", "AAh", "l", "1OW"]
    assert words[0][2] == "hello"
    assert words[0][0] == 2
    assert words[2][1] == ["%"]
    assert words[2][0] == 1
    assert words[-1][1] == ["%"]
    assert len(words) == 8



from saysynth import Segment, Note, MidiTrack, Lyrics, Word


def test_segment():
    segment = Segment(note="A3", velocity=127, duration=100)
    assert segment.to_text() == "[[ volm 1.0 ]] 2m {D 100; P 440:0 440:100}"


def test_note():
    note = Note(
        note="C4",
        phoneme="m",
        velocity=127,
        attack=0.1,
        decay=0.2,
        sustain=0.3,
        release=0.5,
        bpm=126,
        count=1 / 2,
        segment_bpm=126,
        volume_range=[0.1, 0.5],
        segment_count=1 / 32,
    )
    assert (
        note.segments[0].to_text()
        == "[[ volm 0.1 ]] m {D 59.5238; P 523.25:0 523.25:100}"
    )
    assert (
        note.segments[-1].to_text()
        == "[[ volm 0.108 ]] m {D 35.7143; P 523.25:0 523.25:100}"
    )


def test_lyrics():
    lyrics = Lyrics("hello world")
    assert len(lyrics.words) == 2
    assert [isinstance(word, Word) for word in lyrics.words]



"""
Play a sequence of commands concurrently via yaml specification
"""
import copy
from pathos.multiprocessing import ProcessingPool as Pool

import yaml
import click

from saysynth.cli.commands import chord, midi, note, arp


TRACK_FUNCS = {
    "chord": chord.run,
    "midi": midi.run,
    "note": note.run,
    "arp": arp.run,
}

def _run_track_func(item):
    name, kwargs = item
    type = kwargs.get('type', None)
    options = kwargs.get('options', {})
    options['exec'] = True
    if type not in TRACK_FUNCS:
        raise ValueError(f'Invalid track type: {type}. Choose from: {",".join(TRACK_FUNCS.keys())}')
    print(f'Starting track {name} with kwargs: {kwargs}')
    return TRACK_FUNCS.get(type)(**options)

def run(**kwargs):
    base_config = yaml.safe_load(kwargs['base_config'])
    config_overrides = kwargs['config_overrides']
    globals = base_config.pop('globals', {})
    tracks = kwargs.get('tracks') or []
    items = []
    for track_name, base_track_config in base_config.get('tracks', {}).items():

        # optionally skip tracks
        if len(tracks):
            if track_name not in tracks:
                continue

        # allow for track-specific overrides
        track_overrides = config_overrides.pop(track_name, {})
        config_overrides.update(track_overrides)

        # create track config
        track_config = copy.copy(globals) # start with globals
        track_config.update(base_track_config) # override with base track configs
        track_config.update(config_overrides) # override at runtime

        # add to run items
        items.append((track_name, track_config))

    return Pool(len(items)).map(_run_track_func, items)


@click.command()
@click.argument("base_config", type=click.File(), required=True)
@click.option("-t", "--tracks", type=lambda x: [track for track in x.split(',')])
@click.option("-c", "--config-overrides", type=lambda x: yaml.safe_load(x), default='{}', help="Override global and track configurations at runtime")
def cli(**kwargs):
    return run(**kwargs)


import sys

import click

from midi_utils import note_to_midi
from saysynth import say, Arp

from saysynth.cli.options import (
    say_opts,
    start_opts,
    chord_opts,
    phoneme_opts,
    segment_opts,
    adsr_opts,
    arp_opts,
    duration_opts,
    velocity_emphasis_opt,
    volume_range_opt,
    randomize_velocity_opt,
    prepare_options_for_say
)
from saysynth.lib import say

def run(**kwargs):
    """
    Given a note name (or midi note number), stream text required to generate a continuous drone for input to say
    """
    kwargs['start_duration'] = kwargs.get('start', 0)
    text = Arp(**kwargs).to_say_text()

    # handle writing text to file
    output_file = kwargs.get("output_file")
    if output_file:
        with open(output_file, "w") as f:
            f.write(text)

    # if we're not executing say, write text to stdout
    elif not kwargs.get('exec', False):
        sys.stdout.write(text)

    else:
        say.run(**prepare_options_for_say(text, **kwargs))

@click.command()
@click.argument("root", type=note_to_midi, default="A2")
@arp_opts
@chord_opts
@velocity_emphasis_opt
@volume_range_opt
@randomize_velocity_opt
@start_opts
@duration_opts
@phoneme_opts
@segment_opts
@adsr_opts
@click.option(
    "-o",
    "--output-file",
    type=str,
    help="A filepath to write the generated text to",
)
@say_opts
def cli(**kwargs):
    return run(**kwargs)



import os
from typing import Generator, Dict, List, Any

import mido
from midi_utils import midi_to_note


def load(midi_file: str) -> List[mido.Message]:
    """
    Load and validate a midi file, returning a list of messages
    """
    if not os.path.exists(midi_file):
        raise ValueError(f"{midi_file} does not exist")

    mid = mido.MidiFile(midi_file)

    # TODO: handle multitrack midi files
    if len(mid.tracks) > 1:
        raise NotImplementedError(
            "There is not currently support for multi-track midifiles."
        )

    # filter valid messages
    messages = [msg for msg in mid if msg.type in ["note_on", "note_off"]]

    # validate first message
    if messages[0].type != "note_on":
        raise ValueError(
            "This midi file does not start with a note_on message. "
            "Reformat and try again"
        )

    # validate final message
    if messages[-1].type != "note_off":
        raise ValueError(
            "This midi file does not end with a note_off message. "
            "Reformat and try again"
        )
    return messages


def process(midi_file: str) -> Generator[Dict[str, Any], None, None]:
    """
    Load a midi file and yield say-friendly parameters
    """
    messages = load(midi_file)
    total_time = 0
    for i, curr_msg in enumerate(messages):
        total_time += curr_msg.time
        prev_msg = messages[i - 1]

        # ignore first note_on message and only yield when
        # we get a note off message
        if i > 0 and curr_msg.type == "note_off":
            if prev_msg.type != "note_on":
                raise ValueError(
                    f"Overlapping note {midi_to_note(prev_msg.note)} found at message #{i} @ {total_time:.2f}s. "
                    "Only fully monophonic tracks are supported. "
                    "Reformat and try again."
                )
            # if previous note on message has a delta time, yield silence
            if prev_msg.time > 0:
                yield {
                    "type": "silence",
                    "note": None,
                    "duration": prev_msg.time * 1000.0,
                    "velocity": 0,
                }
            # yield the note with computed duration
            yield {
                "type": "note",
                "note": prev_msg.note,
                "duration": curr_msg.time * 1000.0,
                "velocity": prev_msg.velocity,
            }



from setuptools import setup, find_packages

config = {
    "name": "saysynth",
    "version": "0.0.27",
    "packages": find_packages(),
    "install_requires": [
        "charset-normalizer",
        "click",
        "mido",
        "midi-utils",
        "pyyaml",
        "pathos",
        "g2p_en",
        "nltk",
    ],
    "author": "Brian Abelson",
    "author_email": "hey@gltd.email",
    "description": "A synthesizer built on say",
    "url": "http://globally.ltd",
    "entry_points": {
        "console_scripts": ["saysynth=saysynth.cli:main", "sy=saysynth.cli:main"]
    },
}

setup(**config)



from saysynth.utils import here
from saysynth.lib import midi


def test_valid_midi_file():
    messages = list(midi.process(here(__file__, "fixtures/valid.mid")))
    assert messages[0]["type"] == "note"
    assert messages[0]["note"] == 48
    assert messages[0]["velocity"] == 100
    assert messages[0]["duration"] == 4000.0
    assert messages[1]["type"] == "silence"
    assert messages[1]["duration"] == 500.0
    assert messages[-1]["type"] == "note"
    assert messages[-1]["note"] == 44



import copy

import click

from midi_utils import note_to_midi
from saysynth import Chord
from saysynth.cli.options import (
    say_opts,
    velocity_opts,
    phoneme_opts,
    segment_opts,
    chord_opts,
    adsr_opts,
    duration_opts,
    start_opts
)

# # # # #
# Chord #
# # # # #

def run(**kwargs):
    """
    Given a note name (or midi note number), stream text required to generate a continuous drone for input to say
    """
    # generate chord
    chord = Chord(**kwargs)
    # handle writing text to multiple files
    output_file = kwargs.pop("output_file", None)
    if output_file:
        chord.write(output_file)
    else:
        chord.play()


@click.command()
@click.argument("root", type=note_to_midi, default="A2")
@chord_opts
@start_opts
@duration_opts
@velocity_opts
@adsr_opts
@phoneme_opts
@segment_opts
@click.option(
    "-o",
    "--output-file",
    type=str,
    help="A filepath to write the generated text to. This filepath will be used as a pattern to generate multiple text files, one per note in the chord.",
)
@say_opts
def cli(**kwargs):
    return run(**kwargs)